/***************************************************************************
 *                                  _   _ ____  _
 *  Project                     ___| | | |  _ \| |
 *                             / __| | | | |_) | |
 *                            | (__| |_| |  _ <| |___
 *                             \___|\___/|_| \_\_____|
 *
 * Copyright (C) 1998 - 2022, Daniel Stenberg, <daniel@haxx.se>, et al.
 *
 * This software is licensed as described in the file COPYING, which
 * you should have received as part of this distribution. The terms
 * are also available at https://curl.se/docs/copyright.html.
 *
 * You may opt to use, copy, modify, merge, publish, distribute and/or sell
 * copies of the Software, and permit persons to whom the Software is
 * furnished to do so, under the terms of the COPYING file.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 * SPDX-License-Identifier: curl
 *
 ***************************************************************************/

#include "curl_printf.h"

#include <string.h>
#include <locale.h>

#define TOTAL_STR_LEN 4

int main(void)
{
  char zero[TOTAL_STR_LEN] = {'\0'};
  int chars;

  setlocale(LC_NUMERIC, "");
  chars = msnprintf(zero, TOTAL_STR_LEN, "%.1f", 0.0);
  if((chars == (TOTAL_STR_LEN - 1)) && (strcmp(zero, "0.0") == 0))
    return 0;
  else
    return 1;
}
